"""Tutorial module with data for docs and quick testing."""
import numpy as np
import xarray as xr

try:
    from IPython import get_ipython
    from PIL.Image import fromarray

    array_display = True  # pylint: disable=invalid-name
except ModuleNotFoundError:
    array_display = False  # pylint: disable=invalid-name


def generate_mcmc_like_dataset(seed=None):
    """Generate a Dataset with multiple variables, some with dimensions from mcmc sampling.

    Parameters
    ----------
    seed : int or sequence of int, optional
        The random seed used to initialize :func:`numpy.random.default_rng`.

    Examples
    --------
    The dataset generated is the following:

    .. jupyter-execute::

        from xarray_einstats import tutorial
        tutorial.generate_mcmc_like_dataset(3)

    Notes
    -----
    This function is not part of the public API and is designed for use in our documentation.
    In addition to generating the data, it also sets ``display_expand_data=False`` to
    avoid taking too much virtual space with examples.

    """
    xr.set_options(display_expand_data=False)
    rng = np.random.default_rng(seed)
    ds = xr.Dataset(
        {
            "x_plot": (["plot_dim"], np.linspace(0, 10, 20)),
            "mu": (["chain", "draw", "team"], rng.exponential(size=(4, 10, 6))),
            "sigma": (["chain", "draw"], rng.exponential(size=(4, 10))),
            "score": (["chain", "draw", "match"], rng.poisson(size=(4, 10, 12))),
        },
        coords={"team": list("abcdef"), "chain": np.arange(4), "draw": np.arange(10)},
    )
    return ds


def generate_matrices_dataarray(seed=None):
    """Generate a 4d DataArray representing a batch of matrices.

    Parameters
    ----------
    seed : int or sequence of int, optional
        The random seed used to initialize :func:`numpy.random.default_rng`.

    Examples
    --------
    The dataset generated is the following:

    .. jupyter-execute::

        from xarray_einstats import tutorial
        tutorial.generate_matrices_dataarray(5)

    Notes
    -----
    This function is not part of the public API and is designed for use in our documentation.
    In addition to generating the data, it also sets ``display_expand_data=False`` to
    avoid taking too much virtual space with examples.

    """
    xr.set_options(display_expand_data=False)
    rng = np.random.default_rng(seed)
    return xr.DataArray(
        rng.exponential(size=(10, 3, 4, 4)), dims=["batch", "experiment", "dim", "dim2"]
    )


if array_display:

    def display_np_arrays_as_images():
        """Display numpy arrays as images by default in IPython and Jupyter.

        Only needs to be imported for the behaviour to be configured.
        """
        # pylint: disable=protected-access

        def np_to_png(a):
            if 2 <= len(a.shape) <= 3:
                return fromarray(np.array(np.clip(a, 0, 1) * 255, dtype="uint8"))._repr_png_()
            return fromarray(np.zeros([1, 1], dtype="uint8"))._repr_png_()

        def np_to_text(obj, p, cycle):  # pylint: disable=unused-argument, invalid-name
            if len(obj.shape) < 2:
                print(repr(obj))
            if 2 <= len(obj.shape) <= 3:
                pass
            else:
                print(f"<array of shape {obj.shape}>")

        get_ipython().display_formatter.formatters["image/png"].for_type(np.ndarray, np_to_png)
        get_ipython().display_formatter.formatters["text/plain"].for_type(np.ndarray, np_to_text)



"""Stats, linear algebra and einops for xarray."""

from __future__ import annotations

import numpy as np
import xarray as xr

from .linalg import einsum, einsum_path, matmul
from .accessors import LinAlgAccessor, EinopsAccessor

__all__ = [
    "einsum",
    "einsum_path",
    "matmul",
    "zeros_ref",
    "ones_ref",
    "empty_ref",
    "LinAlgAccessor",
    "EinopsAccessor",
]

__version__ = "0.6.0"


def sort(da, dim, **kwargs):
    """Sort along dimension using DataArray values."""
    sort_kwargs = {"axis": -1}
    if "kind" in kwargs:
        sort_kwargs["kind"] = kwargs.pop("kind")
    return xr.apply_ufunc(
        np.sort,
        da,
        input_core_dims=[[dim]],
        output_core_dims=[[dim]],
        kwargs=sort_kwargs,
        **kwargs,
    )


def _remove_indexes_to_reduce(da, dims):
    """Remove indexes related to provided dims.

    Removes indexes related to dims on which we need to operate.
    As many functions only support integer `axis` or None,
    in order to have our functions operate on multiple dimensions
    we need to stack/flatten them. If some of those dimensions
    are already indexed by a multiindex this doesn't work, so we
    remove the indexes. As they are reduced, that information
    will end up being lost eventually either way.
    """
    index_keys = list(da.indexes)
    remove_indicator = [
        (any(da.indexes[k] is index for k in index_keys if k in dims))
        for name, index in da.indexes.items()
    ]
    indexes_to_remove = [k for k, remove in zip(index_keys, remove_indicator) if remove]
    da = da.drop_indexes(indexes_to_remove)
    coords_to_remove = [coord for coord in da.coords if coord in indexes_to_remove or coord in dims]
    return da.reset_coords(coords_to_remove, drop=True)


def _find_index(elem, to_search_in):
    for i, da in enumerate(to_search_in):
        if elem in da.dims:
            return (i, elem)
    raise ValueError(f"{elem} not found in any reference object")


def _create_ref(*args, dims, np_creator, dtype=None):
    if dtype is None and all(isinstance(arg, xr.DataArray) for arg in args):
        dtype = np.result_type(*[arg.dtype for arg in args])
    ref_idxs = [_find_index(dim, args) for dim in dims]
    shape = [len(args[idx][dim]) for idx, dim in ref_idxs]
    # TODO: keep non indexing coords?
    coords = {dim: args[idx][dim] for idx, dim in ref_idxs if dim in args[idx].coords}
    return xr.DataArray(
        np_creator(shape, dtype=dtype),
        dims=dims,
        coords=coords,
    )


def zeros_ref(*args, dims, dtype=None):
    """Create a zeros DataArray from reference object(s).

    Creates a DataArray filled with zeros from reference
    DataArrays or Datasets and a list with the desired dimensions.

    Parameters
    ----------
    *args : iterable of DataArray or Dataset
        Reference objects from which the lengths and coordinate values (if any)
        of the given `dims` will be taken.
    dims : list of hashable
        List of dimensions of the output DataArray. Passed as is to the
        {class}`~xarray.DataArray` constructor.
    dtype : dtype, optional
        The dtype of the output array.
        If it is not provided it will be inferred from the reference
        DataArrays in `args` with :func:`numpy.result_type`.

    Returns
    -------
    DataArray

    See Also
    --------
    ones_ref, empty_ref
    """
    return _create_ref(*args, dims=dims, np_creator=np.zeros, dtype=dtype)


def empty_ref(*args, dims, dtype=None):
    """Create an empty DataArray from reference object(s).

    Creates an empty DataArray from reference
    DataArrays or Datasets and a list with the desired dimensions.

    Parameters
    ----------
    *args : iterable of DataArray or Dataset
        Reference objects from which the lengths and coordinate values (if any)
        of the given `dims` will be taken.
    dims : list of hashable
        List of dimensions of the output DataArray. Passed as is to the
        {class}`~xarray.DataArray` constructor.
    dtype : dtype, optional
        The dtype of the output array.
        If it is not provided it will be inferred from the reference
        DataArrays in `args` with :func:`numpy.result_type`.

    Returns
    -------
    DataArray

    See Also
    --------
    ones_ref, zeros_ref
    """
    return _create_ref(*args, dims=dims, np_creator=np.empty, dtype=dtype)


def ones_ref(*args, dims, dtype=None):
    """Create a ones DataArray from reference object(s).

    Creates a DataArray filled with ones from reference
    DataArrays or Datasets and a list with the desired dimensions.

    Parameters
    ----------
    *args : iterable of DataArray or Dataset
        Reference objects from which the lengths and coordinate values (if any)
        of the given `dims` will be taken.
    dims : list of hashable
        List of dimensions of the output DataArray. Passed as is to the
        {class}`~xarray.DataArray` constructor.
    dtype : dtype, optional
        The dtype of the output array.
        If it is not provided it will be inferred from the reference
        DataArrays in `args` with :func:`numpy.result_type`.

    Returns
    -------
    DataArray

    See Also
    --------
    empty_ref, zeros_ref
    """
    return _create_ref(*args, dims=dims, np_creator=np.ones, dtype=dtype)



"""Wrappers for :mod:`numpy.linalg`.

.. tip::

    Most of the functions in this module are also available via the ``.linalg`` accessor
    from :class:`DataArray` objects. See :ref:`linalg_tutorial` for
    example usage.

    The functions that are not available via the accessor are ``einsum``, ``einsum_path``,
    ``matmul`` and ``get_default_dims``.

"""
import warnings

import numpy as np
import xarray as xr

__all__ = [
    "matrix_power",
    "matrix_transpose",
    "cholesky",
    "qr",
    "svd",
    "eig",
    "eigh",
    "eigvals",
    "eigvalsh",
    "norm",
    "cond",
    "det",
    "matrix_rank",
    "slogdet",
    "trace",
    "diagonal",
    "solve",
    "inv",
]


class MissingMonkeypatchError(Exception):
    """Error specific for the linalg module non-default yet accepted monkeypatch."""


def get_default_dims(da1_dims, d2_dims=None):
    """Get the dimensions corresponding to the matrices.

    Parameters
    ----------
    da1_dims : list of str

    da2_dims : list of str, optional
        Used only in case of multiple inputs, otherwise it will keep its default value of ``None``

    Returns
    -------
    list of str
        The dimensions indicating the matrix dimensions. Must be an iterable
        containing two strings.

    Warnings
    --------
    ``dims`` is required for functions in the linalg module.
    This function acts as a placeholder and only raises an error indicating
    that dims is a required argument unless this function is monkeypatched.

    It is documented here to show how to write and configure a substitute function.

    Examples
    --------
    The ``xarray_einstats`` default behaviour is requiring the `dims` argument
    for functions in the linalg module. Not providing it raises a `TypeError`

    .. jupyter-execute::
        :raises: TypeError

        from xarray_einstats import linalg, tutorial
        da = tutorial.generate_matrices_dataarray(5)
        linalg.inv(da)

    You need to pass the dimensions corresponding the matrix axes explicitly

    .. jupyter-execute::

        linalg.inv(da, dims=["dim", "dim2"])

    However, in many cases it will be possible to identify those dimensions
    from the list of all dimension names in the input.

    Here we show how to monkeypatch ``get_default_dims`` to get a different default
    behaviour. If you follow a convention to label the dimensions corresponding
    to the matrix axes, you can integrate this logic into ``xarray_einstats``,
    which will avoid unnecessary repetition, especially if performing several
    chained linear algebra operations:

    .. jupyter-execute::

        def get_default_dims(dims1, dims2):
            if dims2 is not None:
                raise TypeError("Default dims only valid for single input functions")
            matrix_dims = [dim for dim in dims1 if f"{dim}2" in dims1]
            if len(matrix_dims) != 1:
                raise TypeError("Unable to guess default matrix dims")
            dim = matrix_dims[0]
            return [dim, f"{dim}2"]

        linalg.get_default_dims = get_default_dims
        linalg.inv(da)

    You can still use ``dims`` explicitly to override those defaults.

    """
    raise MissingMonkeypatchError()


def _attempt_default_dims(func, da1_dims, da2_dims=None):
    """Raise a more informative warning."""
    try:
        aux = get_default_dims(da1_dims, da2_dims)
    except MissingMonkeypatchError:
        raise TypeError(
            f"{func} missing required argument dims. You must monkeypatch "
            "xarray_einstats.linalg.get_default_dims for dims=None to be supported"
        ) from None
    return aux


class PairHandler:
    def __init__(self, all_dims, keep_dims):
        self.potential_out_dims = keep_dims.union(all_dims)
        self.einsum_axes = list(
            letter
            for letter in "zyxwvutsrqponmlkjihgfedcba"
            if letter not in self.potential_out_dims
        )
        self.dim_map = {d: self.einsum_axes.pop() for d in all_dims}
        self.out_dims = []
        self.out_subscript = ""

    def process_dim_da_pair(self, da, dim_sublist):
        da_dims = da.dims
        out_dims = [
            dim for dim in da_dims if dim in self.potential_out_dims and dim not in dim_sublist
        ]
        subscripts = ""
        updated_in_dims = dim_sublist.copy()
        for dim in out_dims:
            self.out_dims.append(dim)
            sub = self.einsum_axes.pop()
            self.out_subscript += sub
            subscripts += sub
            updated_in_dims.insert(0, dim)
        for dim in dim_sublist:
            subscripts += self.dim_map[dim]
        if len(da_dims) > len(out_dims) + len(dim_sublist):
            return f"...{subscripts}", updated_in_dims
        return subscripts, updated_in_dims

    def get_out_subscript(self):
        if not self.out_subscript:
            return ""
        return f"->{self.out_subscript}"


def _einsum_parent(dims, *operands, keep_dims=frozenset()):
    """Preprocess inputs to call :func:`numpy.einsum` or :func:`numpy.einsum_path`.

    Parameters
    ----------
    dims : list of list of str
        List of lists of dimension names. It must have the same length or be
        only one item longer than ``operands``. If both have the same
        length, the generated pattern passed to {func}`numpy.einsum`
        won't have ``->`` nor right hand side. Otherwise, the last
        item is assumed to be the dimension specification of the output
        DataArray, and it can be an empty list to add ``->`` but no
        subscripts.
    operands : DataArray
        DataArrays for the operation. Multiple DataArrays are accepted.
    keep_dims : set, optional
        Dimensions to exclude from summation unless specifically specified in ``dims``

    See Also
    --------
    xarray_einstats.einsum, xarray_einstats.einsum_path
    numpy.einsum, numpy.einsum_path
    xarray_einstats.einops.reduce
    """
    if len(dims) == len(operands):
        in_dims = dims
        out_dims = None
    elif len(dims) == len(operands) + 1:
        in_dims = dims[:-1]
        out_dims = dims[-1]
    else:
        raise ValueError("length of dims and operands not compatible")

    all_dims = set(dim for sublist in dims for dim in sublist)
    handler = PairHandler(all_dims, keep_dims)
    in_subscripts = []
    updated_in_dims = []
    for da, sublist in zip(operands, in_dims):
        in_subs, up_dims = handler.process_dim_da_pair(da, sublist)
        in_subscripts.append(in_subs)
        updated_in_dims.append(up_dims)

    in_subscript = ",".join(in_subscripts)
    if out_dims is None:
        out_subscript = handler.get_out_subscript()
        out_dims = handler.out_dims
    elif not out_dims:
        out_subscript = "->"
    else:
        out_subscript = "->" + "".join(handler.dim_map[dim] for dim in out_dims)
    if out_subscript and "..." in in_subscript:
        out_subscript = "->..." + out_subscript[2:]
    subscripts = in_subscript + out_subscript
    return subscripts, updated_in_dims, out_dims


def _translate_pattern_string(subscripts):
    """Translate a pattern given as string of dimension names to list of dimension names."""
    if "->" in subscripts:
        in_subscripts, out_subscript = subscripts.split("->")
    else:
        in_subscripts = subscripts
        out_subscript = None
    in_dims = [
        [dim.strip(", ") for dim in in_subscript.split(" ")]
        for in_subscript in in_subscripts.split(",")
    ]
    if out_subscript is None:
        dims = in_dims
    elif not out_subscript:
        dims = [*in_dims, []]
    else:
        out_dims = [dim.strip(", ") for dim in out_subscript.split(" ")]
        dims = [*in_dims, out_dims]
    return dims


def _einsum_path(dims, *operands, keep_dims=frozenset(), optimize=None, **kwargs):
    """Wrap :func:`numpy.einsum_path` directly."""
    op_kwargs = {} if optimize is None else {"optimize": optimize}

    subscripts, in_dims, _ = _einsum_parent(dims, *operands, keep_dims=keep_dims)
    updated_in_dims = []
    for sublist, da in zip(in_dims, operands):
        updated_in_dims.append([dim for dim in da.dims if dim not in sublist] + sublist)

    return xr.apply_ufunc(
        np.einsum_path,
        subscripts,
        *operands,
        input_core_dims=[[], *updated_in_dims],
        output_core_dims=[[]],
        kwargs=op_kwargs,
        **kwargs,
    ).values.item()


def einsum_path(dims, *operands, keep_dims=frozenset(), optimize=None, **kwargs):
    """Expose :func:`numpy.einsum_path` with an xarray-like API.

    See :func:`xarray_einstats.einsum` for a detailed description of ``dims``
    and ``operands``.

    Parameters
    ----------
    dims : list of list of str
    operands : DataArray
    optimize : str, optional
        ``optimize`` argument for :func:`numpy.einsum_path`. It defaults to None so that
        we always default to numpy's default, without needing to keep the call signature
        here up to date.
    kwargs : dict, optional
        Passed to :func:`xarray.apply_ufunc`
    """
    if isinstance(dims, str):
        dims = _translate_pattern_string(dims)
    return _einsum_path(dims, *operands, keep_dims=keep_dims, optimize=optimize, **kwargs)


def _einsum(dims, *operands, keep_dims=frozenset(), out_append="{i}", einsum_kwargs=None, **kwargs):
    """Wrap :func:`numpy.einsum` directly.

    The user facing version is :func:`xarray_einstats.einsum` which exposes two APIs.
    """
    if einsum_kwargs is None:
        einsum_kwargs = {}

    subscripts, updated_in_dims, out_dims = _einsum_parent(dims, *operands, keep_dims=keep_dims)

    updated_out_dims = []
    for i, dim in enumerate(out_dims):
        totalcount = out_dims.count(dim)
        count = out_dims[:i].count(dim) + 1
        updated_out_dims.append(
            dim + out_append.format(i=count) if (totalcount > 1) and (count > 1) else dim
        )
    return xr.apply_ufunc(
        np.einsum,
        subscripts,
        *operands,
        input_core_dims=[[], *updated_in_dims],
        output_core_dims=[updated_out_dims],
        kwargs=einsum_kwargs,
        **kwargs,
    )


def raw_einsum(*args, **kwargs):
    """Wrap numpy.einsum.

    DEPRECATED
    """
    warnings.warn(
        "`raw_einsum` has been deprecated. Its functionality has been merged into `einsum`",
        DeprecationWarning,
    )
    return einsum(*args, **kwargs)


def einsum(dims, *operands, keep_dims=frozenset(), out_append="{i}", einsum_kwargs=None, **kwargs):
    """Expose :func:`numpy.einsum` with an xarray-like API.

    Usage examples of all arguments is available at the
    :ref:`einsum section <linalg_tutorial/einsum>` of the linear algebra module tutorial.

    Parameters
    ----------
    dims : str or list of list of str
        If `dims` is a string it is intepreted as the subscripts for the summation as dimension
        names. Spaces indicate multiple dimensions in a DataArray and commas indicate
        multiple DataArray operands.
        Only dimensions with no spaces, nor commas nor ``->`` characters are valid.

        If `dims` is a list it is interpreted as list of lists of dimension names.
        It must have the same length or be only one item longer than `operands`.
        If both have the same length, the generated pattern passed to {func}`numpy.einsum`
        won't have ``->`` nor right hand side. Otherwise, the last
        item is assumed to be the dimension specification of the output
        DataArray. In this case it can be an empty list to add ``->`` but no
        subscripts.
    operands : DataArray
        DataArrays for the operation. Multiple DataArrays are accepted.
    keep_dims : set, optional
        Dimensions to exclude from summation unless specifically specified in ``dims``
    out_append : str, optional
        Pattern to append to repeated dimension names in the output (if any). The pattern should
        contain a substitution for variable ``i``, which indicates the number of the current
        dimension among the repeated ones. Its default value is ``"{i}"``.
        To keep repeated dimension names use ``""``.

        The first occurrence will keep the original name and not use ``out_append``.
        It will therefore inherit the coordinate values in case there were any.
    einsum_kwargs : dict, optional
        Passed to :func:`numpy.einsum`
    kwargs : dict, optional
        Passed to :func:`xarray.apply_ufunc`

    Notes
    -----
    Dimensions present in ``dims`` will be reduced, but unlike {func}`xarray.dot` it does so only
    for that variable.
    """
    if isinstance(dims, str):
        dims = _translate_pattern_string(dims)
    return _einsum(
        dims,
        *operands,
        keep_dims=keep_dims,
        out_append=out_append,
        einsum_kwargs=einsum_kwargs,
        **kwargs,
    )


def matmul(da, db, dims=None, *, out_append="2", **kwargs):
    """Wrap :func:`numpy.linalg.matmul`.

    Usage examples of all arguments is available at the
    :ref:`matmul section <linalg_tutorial/matmul>` of the linear algebra module tutorial.
    """
    rename = False
    if dims is None:
        dims = _attempt_default_dims("matmul", da.dims, db.dims)
    if len(dims) == 3:
        dim1, dim2, dim3 = dims
        dims1 = [dim1, dim2]
        dims2 = [dim2, dim3]
        out_dims = [dim1, dim3]
        if dim1 == dim3:
            db = db.rename({dim3: dim3 + out_append})
            dims2 = [dim2, dim3 + out_append]
            out_dims = [dim1, dim3 + out_append]
        else:
            if dim3 in da.dims:
                da = da.rename({dim3: dim3 + out_append})
            if dim1 in db.dims:
                db = db.rename({dim1: dim1 + out_append})
    elif len(dims) != 2:
        raise ValueError(
            "matmul can be one of '[str, str]', '[str, str, str]' or '[[str, str], [str, str]]'"
        )
    elif isinstance(dims[0], str):
        dims1 = dims
        dims2 = dims
        out_dims = dims
    else:
        rename = True
        dim11, dim12 = dims[0]
        dim21, dim22 = dims[1]
        da = da.rename({dim11: "__aux_dim11__", dim12: "__aux_dim12__"})
        db = db.rename({dim21: "__aux_dim21__", dim22: "__aux_dim22__"})
        dims1 = ["__aux_dim11__", "__aux_dim12__"]
        dims2 = ["__aux_dim21__", "__aux_dim22__"]
        out_dims = ["__aux_dim11__", "__aux_dim22__"]
    matmul_aux = xr.apply_ufunc(
        np.matmul,
        da,
        db,
        input_core_dims=[dims1, dims2],
        output_core_dims=[out_dims],
        **kwargs,
    )
    if rename:
        return matmul_aux.rename(
            __aux_dim11__=dim11, __aux_dim22__=dim22 + out_append if dim22 == dim11 else dim22
        )
    return matmul_aux


def matrix_transpose(da, dims):
    """Transpose the underlying matrix without modifying the dimensions.

    This convenience function uses :meth:`~xarray.DataArray.swap_dims` followed
    by :meth:`~xarray.DataArray.transpose` to get the equivalent of a matrix transposition.

    Parameters
    ----------
    da : DataArray
        Input DataArray
    dims : list of str
        Matrix dimensions

    Returns
    -------
    DataArray
        The DataArray after transposing the matrix data but leaving the dimensions untouched.
    """
    if dims is None:
        dims = _attempt_default_dims("matrix_power", da.dims)
    dim1, dim2 = dims
    return da.swap_dims({dim1: dim2, dim2: dim1}).transpose(..., *dims)


def matrix_power(da, n, dims=None, **kwargs):
    """Wrap :func:`numpy.linalg.matrix_power`.

    Usage examples of all arguments is available at the :ref:`linalg_tutorial` page.
    """
    if dims is None:
        dims = _attempt_default_dims("matrix_power", da.dims)
    return xr.apply_ufunc(
        np.linalg.matrix_power, da, n, input_core_dims=[dims, []], output_core_dims=[dims], **kwargs
    )


def cholesky(da, dims=None, **kwargs):
    """Wrap :func:`numpy.linalg.cholesky`.

    Usage examples of all arguments is available at the :ref:`linalg_tutorial` page.
    """
    if dims is None:
        dims = _attempt_default_dims("cholesky", da.dims)
    return xr.apply_ufunc(
        np.linalg.cholesky, da, input_core_dims=[dims], output_core_dims=[dims], **kwargs
    )


def qr(da, dims=None, *, mode="reduced", out_append="2", **kwargs):  # pragma: no cover
    """Wrap :func:`numpy.linalg.qr`.

    Usage examples of all arguments is available at the :ref:`linalg_tutorial` page.
    """
    if dims is None:
        dims = _attempt_default_dims("qr", da.dims)
    m_dim, n_dim = dims
    m, n = len(da[m_dim]), len(da[n_dim])
    k, k_dim = (m, m_dim) if n >= m else (n, n_dim)
    mode = mode.lower()
    if mode == "reduced":
        out_dims = [
            [m_dim, k_dim + (out_append if k_dim == m_dim else "")],
            [k_dim, n_dim + (out_append if k_dim == n_dim else "")],
        ]
    elif mode == "complete":
        out_dims = [[m_dim, m_dim + out_append], [m_dim, n_dim]]
    elif mode == "r":
        out_dims = [[m_dim if k == m else n_dim + out_append, n_dim]]
    elif mode == "raw":
        out_dims = [[n_dim, m_dim], [m_dim if k == m else n_dim]]
    else:
        raise ValueError("mode not recognized")

    return xr.apply_ufunc(
        np.linalg.qr,
        da,
        input_core_dims=[dims],
        output_core_dims=out_dims,
        kwargs={"mode": mode},
        **kwargs,
    )


def svd(
    da, dims=None, *, full_matrices=True, compute_uv=True, hermitian=False, out_append="2", **kwargs
):
    """Wrap :func:`numpy.linalg.svd`.

    Usage examples of all arguments is available at the :ref:`linalg_tutorial` page.
    """
    if dims is None:
        dims = _attempt_default_dims("svd", da.dims)
    m_dim, n_dim = dims
    m, n = len(da[m_dim]), len(da[n_dim])
    k, k_dim = (m, m_dim) if m <= n else (n, n_dim)
    s_dims = [k_dim]
    if full_matrices:
        u_dims = [m_dim, m_dim + out_append]
        vh_dims = [n_dim, n_dim + out_append]
    else:
        if m == k:
            u_dims = [m_dim, k_dim + out_append]
            vh_dims = [k_dim, n_dim]
        else:
            u_dims = [m_dim, k_dim]
            vh_dims = [k_dim, n_dim + out_append]
    if compute_uv:
        out_dims = [u_dims, s_dims, vh_dims]
    else:
        out_dims = [s_dims]
    return xr.apply_ufunc(
        np.linalg.svd,
        da,
        input_core_dims=[dims],
        output_core_dims=out_dims,
        kwargs={"full_matrices": full_matrices, "compute_uv": compute_uv, "hermitian": hermitian},
        **kwargs,
    )


def eig(da, dims=None, **kwargs):
    """Wrap :func:`numpy.linalg.eig`.

    Usage examples of all arguments is available at the :ref:`linalg_tutorial` page.
    """
    if dims is None:
        dims = _attempt_default_dims("eig", da.dims)
    return xr.apply_ufunc(
        np.linalg.eig, da, input_core_dims=[dims], output_core_dims=[dims[-1:], dims], **kwargs
    )


def eigh(da, dims=None, *, UPLO="L", **kwargs):  # pylint: disable=invalid-name
    """Wrap :func:`numpy.linalg.eigh`.

    Usage examples of all arguments is available at the :ref:`linalg_tutorial` page.
    """
    if dims is None:
        dims = _attempt_default_dims("eigh", da.dims)
    return xr.apply_ufunc(
        np.linalg.eigh,
        da,
        input_core_dims=[dims],
        output_core_dims=[dims[-1:], dims],
        kwargs={"UPLO": UPLO},
        **kwargs,
    )


def eigvals(da, dims=None, **kwargs):
    """Wrap :func:`numpy.linalg.eigvals`.

    Usage examples of all arguments is available at the :ref:`linalg_tutorial` page.
    """
    if dims is None:
        dims = _attempt_default_dims("eigvals", da.dims)
    return xr.apply_ufunc(
        np.linalg.eigvals, da, input_core_dims=[dims], output_core_dims=[dims[-1:]], **kwargs
    )


def eigvalsh(da, dims=None, *, UPLO="L", **kwargs):  # pylint: disable=invalid-name
    """Wrap :func:`numpy.linalg.eigvalsh`.

    Usage examples of all arguments is available at the :ref:`linalg_tutorial` page.
    """
    if dims is None:
        dims = _attempt_default_dims("eigvalsh", da.dims)
    return xr.apply_ufunc(
        np.linalg.eigvalsh,
        da,
        input_core_dims=[dims],
        output_core_dims=[dims[-1:]],
        kwargs={"UPLO": UPLO},
        **kwargs,
    )


def norm(da, dims=None, *, ord=None, **kwargs):  # pylint: disable=redefined-builtin
    """Wrap :func:`numpy.linalg.norm`.

    Usage examples of all arguments is available at the :ref:`linalg_tutorial` page.
    """
    if dims is None:
        dims = _attempt_default_dims("norm", da.dims)
    norm_kwargs = {"ord": ord}
    if isinstance(dims, str):
        in_dims = [dims]
        norm_kwargs["axis"] = -1
    else:
        in_dims = dims
        norm_kwargs["axis"] = (-2, -1)
    return xr.apply_ufunc(
        np.linalg.norm, da, input_core_dims=[in_dims], kwargs=norm_kwargs, **kwargs
    )


def cond(da, dims=None, *, p=None, **kwargs):  # pylint: disable=invalid-name
    """Wrap :func:`numpy.linalg.cond`.

    Usage examples of all arguments is available at the :ref:`linalg_tutorial` page.
    """
    if dims is None:
        dims = _attempt_default_dims("cond", da.dims)
    return xr.apply_ufunc(np.linalg.cond, da, input_core_dims=[dims], kwargs={"p": p}, **kwargs)


def det(da, dims=None, **kwargs):
    """Wrap :func:`numpy.linalg.det`.

    Usage examples of all arguments is available at the :ref:`linalg_tutorial` page.
    """
    if dims is None:
        dims = _attempt_default_dims("det", da.dims)
    return xr.apply_ufunc(np.linalg.det, da, input_core_dims=[dims], **kwargs)


def matrix_rank(da, dims=None, *, tol=None, hermitian=False, **kwargs):
    """Wrap :func:`numpy.linalg.matrix_rank`.

    Usage examples of all arguments is available at the :ref:`linalg_tutorial` page.
    """
    if dims is None:
        dims = _attempt_default_dims("matrix_rank", da.dims)
    return xr.apply_ufunc(
        np.linalg.matrix_rank,
        da,
        input_core_dims=[dims],
        kwargs={"tol": tol, "hermitian": hermitian},
        **kwargs,
    )


def slogdet(da, dims=None, **kwargs):
    """Wrap :func:`numpy.linalg.slogdet`.

    Usage examples of all arguments is available at the :ref:`linalg_tutorial` page.
    """
    if dims is None:
        dims = _attempt_default_dims("slogdet", da.dims)
    return xr.apply_ufunc(
        np.linalg.slogdet, da, input_core_dims=[dims], output_core_dims=[[], []], **kwargs
    )


def trace(da, dims=None, *, offset=0, dtype=None, out=None, **kwargs):
    """Wrap :func:`numpy.trace`.

    Usage examples of all arguments is available at the :ref:`linalg_tutorial` page.
    """
    if dims is None:
        dims = _attempt_default_dims("trace", da.dims)
    trace_kwargs = {"offset": offset, "dtype": dtype, "out": out, "axis1": -2, "axis2": -1}
    return xr.apply_ufunc(np.trace, da, input_core_dims=[dims], kwargs=trace_kwargs, **kwargs)


def diagonal(da, dims=None, *, offset=0, **kwargs):
    """Wrap :func:`numpy.diagonal`.

    Usage examples of all arguments is available at the :ref:`linalg_tutorial` page.
    """
    if dims is None:
        dims = _attempt_default_dims("diagonal", da.dims)
    diagonal_kwargs = {"offset": offset, "axis1": -2, "axis2": -1}
    out_dims = [dims[0] if offset == 0 else "diag_id"]
    return xr.apply_ufunc(
        np.diagonal,
        da,
        input_core_dims=[dims],
        output_core_dims=[out_dims],
        kwargs=diagonal_kwargs,
        **kwargs,
    )


def solve(da, db, dims=None, **kwargs):
    """Wrap :func:`numpy.linalg.solve`.

    Usage examples of all arguments is available at the :ref:`linalg_tutorial` page.
    """
    if dims is None:
        dims = _attempt_default_dims("solve", da.dims, db.dims)
    if len(dims) == 3:
        b_dim = dims[0] if dims[0] in db.dims else dims[1]
        in_dims = [dims[:2], [b_dim, dims[-1]]]
        out_dims = [[b_dim, dims[-1]]]
    else:
        in_dims = [dims, dims[:1]]
        out_dims = [dims[:1]]
    return xr.apply_ufunc(
        np.linalg.solve, da, db, input_core_dims=in_dims, output_core_dims=out_dims, **kwargs
    )


def inv(da, dims=None, **kwargs):
    """Wrap :func:`numpy.linalg.inv`.

    Usage examples of all arguments is available at the :ref:`linalg_tutorial` page.
    """
    if dims is None:
        dims = _attempt_default_dims("inv", da.dims)
    return xr.apply_ufunc(
        np.linalg.inv, da, input_core_dims=[dims], output_core_dims=[dims], **kwargs
    )



"""Wrappers for `einops <https://einops.rocks/>`_.

The einops module is available only from ``xarray_einstats.einops`` and is not
imported when doing ``import xarray_einstats``.
To use it you need to have installed einops manually or alternatively
install this library as ``xarray-einstats[einops]`` or ``xarray-einstats[all]``.
Details about the exact command are available at :ref:`installation`

.. tip::

    The functions above are also available via the ``.einops`` accessor
    from :class:`DataArray` objects. See :ref:`einops_tutorial` for
    example usage.

"""
import warnings

import einops
import xarray as xr

__all__ = ["rearrange", "reduce", "DaskBackend"]


class DimHandler:
    """Handle converting actual dimension names to placeholders for einops."""

    def __init__(self):
        self.mapping = {}

    def get_name(self, dim):
        """Return or generate a placeholder for a dimension name."""
        if dim in self.mapping:
            return self.mapping.get(dim)
        dim_txt = f"d{len(self.mapping)}"
        self.mapping[dim] = dim_txt
        return dim_txt

    def get_names(self, dim_list):
        """Automate calling get_name with an iterable."""
        return " ".join((self.get_name(dim) for dim in dim_list))

    def rename_kwarg(self, key):
        """Process kwargs for axes_lengths.

        Users use as keys the dimension names they used in the input expressions
        which need to be converted and use the placeholder as key when passed
        to einops functions.
        """
        return self.mapping.get(key, key)


def process_pattern_list(redims, handler, allow_dict=True, allow_list=True):
    """Process a pattern list and convert it to an einops expression using placeholders.

    Parameters
    ----------
    redims : pattern_list
        One of ``out_dims`` or ``in_dims`` in {func}`~xarray_einstats.einops.rearrange`
        or {func}`~xarray_einstats.einops.reduce`.
    handler : DimHandler
    allow_dict, allow_list : bool, optional
        Whether or not to allow lists or dicts as elements of ``redims``.
        When processing ``in_dims`` for example we need the names of
        the variables to be decomposed so dicts are required and lists/tuples
        are not accepted.

    Returns
    -------
    expression_dims : list of str
        A list with the names of the dimensions present in the out expression
    output_dims : list of str
        A list with the names of the dimensions present in the output.
        It differs from ``expression_dims`` because there might be dimensions
        being stacked.
    pattern : str
        The einops expression equivalent to the operations in ``redims`` pattern
        list.

    Examples
    --------
    Whenever we have groupings of dimensions (be it to decompose or to stack),
    ``expression_dims`` and ``output_dims`` differ:

    .. jupyter-execute::

        from xarray_einstats.einops import process_pattern_list, DimHandler
        handler = DimHandler()
        process_pattern_list(["a", {"b": ("c", "d")}, ("e", "f", "g")], handler)

    """
    out = []
    out_names = []
    txt = []
    for subitem in redims:
        if isinstance(subitem, str):
            out.append(subitem)
            out_names.append(subitem)
            txt.append(handler.get_name(subitem))
        elif isinstance(subitem, dict) and allow_dict:
            if len(subitem) != 1:
                raise ValueError(
                    "dicts in pattern list must have a single key but instead "
                    f"found {len(subitem)}: {subitem.keys()}"
                )
            key, values = list(subitem.items())[0]
            if isinstance(values, str):
                raise ValueError("Found values of type str in a pattern dict, use xarray.rename")
            out.extend(values)
            out_names.append(key)
            txt.append(f"( {handler.get_names(values)} )")
        elif allow_list:
            out.extend(subitem)
            out_names.append("-".join(subitem))
            txt.append(f"( {handler.get_names(subitem)} )")
        else:
            raise ValueError(
                f"Found unsupported pattern type: {type(subitem)}, double check the docs. "
                "This could be for example is using lists/tuples as elements of in_dims argument"
            )
    return out, out_names, " ".join(txt)


def translate_pattern(pattern):
    """Translate a string pattern to a list pattern.

    Parameters
    ----------
    pattern : str
        Input pattern as a string.

    Returns
    -------
    pattern_list
        Pattern translated to list, as used by the direct feature-full wrappers.

    Examples
    --------
    .. jupyter-execute::

        from xarray_einstats.einops import translate_pattern
        translate_pattern("a (c d)=b (e f g)")

    """
    dims = []
    current_dim = ""
    current_block = []
    parsing_block = 0  # 0=no block, 1=block, 2=just closed, waiting for key
    parsing_key = False
    for char in pattern.strip() + " ":
        if char == " ":
            if parsing_key:
                if current_dim:
                    dims.append({current_dim: current_block})
                else:
                    dims.append(current_block)
                current_block = []
                parsing_key = False
                parsing_block = False
            elif not current_dim:
                continue
            elif parsing_block:
                current_block.append(current_dim)
            else:
                dims.append(current_dim)
            current_dim = ""
        elif char == ")":
            if parsing_block:
                parsing_block = False
                parsing_key = True
                if current_dim:
                    current_block.append(current_dim)
                current_dim = ""
            else:
                raise ValueError("unmatched parenthesis")
        elif char == "(":
            parsing_block = 1
        elif char == "=":
            if not parsing_key:
                raise ValueError("= sign must follow a closing parenthesis )")
        else:
            current_dim += char
    return dims


def _rearrange(da, out_dims, in_dims=None, **kwargs):
    """Wrap `einops.rearrange <https://einops.rocks/api/rearrange/>`_.

    This is the function that actually interfaces with ``einops``.
    :func:`xarray_einstats.einops.rearrange` is the user facing
    version as it exposes two possible APIs, one of them significantly
    less verbose and more friendly (but much less flexible).

    Parameters
    ----------
    da : xarray.DataArray
        Input DataArray to be rearranged
    out_dims : list of str, list or dict
        See docstring of :func:`~xarray_einstats.einops.rearrange`
    in_dims : list of str or dict, optional
        See docstring of :func:`~xarray_einstats.einops.rearrange`
    kwargs : dict, optional
        kwargs with key equal to dimension names in ``out_dims``
        (that is, strings or dict keys) are passed to einops.rearrange
        the rest of keys are passed to :func:`xarray.apply_ufunc`
    """
    da_dims = da.dims

    handler = DimHandler()
    if in_dims is None:
        in_dims = []
        in_names = []
        in_pattern = ""
    else:
        in_dims, in_names, in_pattern = process_pattern_list(
            in_dims, handler=handler, allow_list=False
        )
    # note, not using sets for da_dims to avoid transpositions on missing variables,
    # if they wanted to transpose those they would not be missing variables
    out_dims, out_names, out_pattern = process_pattern_list(out_dims, handler=handler)
    missing_in_dims = [dim for dim in da_dims if dim not in in_names]
    expected_missing = set(out_dims).union(in_names).difference(in_dims)
    missing_out_dims = [dim for dim in da_dims if dim not in expected_missing]

    # avoid using dimensions as core dims unnecesarly
    non_core_dims = [dim for dim in missing_in_dims if dim in missing_out_dims]
    missing_in_dims = [dim for dim in missing_in_dims if dim not in non_core_dims]
    missing_out_dims = [dim for dim in missing_out_dims if dim not in non_core_dims]

    non_core_pattern = handler.get_names(non_core_dims)
    pattern = f"{non_core_pattern} {handler.get_names(missing_in_dims)} {in_pattern} ->\
        {non_core_pattern} {handler.get_names(missing_out_dims)} {out_pattern}"

    axes_lengths = {
        handler.rename_kwarg(k): v for k, v in kwargs.items() if k in out_names + out_dims
    }
    kwargs = {k: v for k, v in kwargs.items() if k not in out_names + out_dims}
    return xr.apply_ufunc(
        einops.rearrange,
        da,
        pattern,
        input_core_dims=[missing_in_dims + in_names, []],
        output_core_dims=[missing_out_dims + out_names],
        kwargs=axes_lengths,
        **kwargs,
    )


def rearrange(da, pattern, pattern_in=None, **kwargs):
    """Expose `einops.rearrange <https://einops.rocks/api/rearrange/>`_ with an xarray-like API.

    It has two possible syntaxes which are independent and somewhat complementary.

    Wrapper around einops.rearrange with a very similar syntax.
    Spaces, parenthesis ``()`` and `->` are not allowed in dimension names.

    Parameters
    ----------
    da : xarray.DataArray
        Input array
    pattern : str or list of [hashable, list or dict]
        If `pattern` is a string, it uses the same syntax as einops
        with two caveats:

        * Unless splitting or stacking, you must use the actual dimension names.
        * When splitting or stacking you can use ``(dim1 dim2)=dim``. This is *necessary*
          for the left hand side as it identifies the dimension to split, and
          optional on the right hand side, if omitted the stacked dimension will be given
          a default name.

        If `pattern` is not a string, then it must be a list where each of its elements
        is one of: ``str``, ``list`` (to stack those dimensions and give them an
        arbitrary name) or ``dict of {str: list}`` (to stack the dimensions indicated
        as values of the dictionary and name the resulting dimensions with the key).

        `pattern` is then interpreted as the output side of the einops pattern. See
        TODO for more details.
    pattern_in : list of [str or dict], optional
        The input pattern for the dimensions. It can only be provided if `pattern`
        is a ``list``. Also, note this is only necessary if you want to split some dimensions.

        The syntax and interpretation is the same as the case when `pattern` is a list,
        with the only difference that ``list`` elements are not allowed, the same way
        that ``(dim1 dim2)=dim`` is required on the left hand side when using string
        patterns.
    kwargs : dict, optional
        Passed to :func:`xarray_einstats.einops.rearrange`

    Returns
    -------
    xarray.DataArray

    Notes
    -----
    Unlike for general xarray objects, where dimension
    names can be :term:`hashable <xarray:name>` here
    dimension names are not recommended but required to be
    strings for both cases. Future releases however might
    support this when using lists as `pattern`, comment
    on :issue:`50` if you are interested in the feature
    or could help implement it.


    See Also
    --------
    xarray_einstats.einops.reduce
    """
    if isinstance(pattern, str):
        if "->" in pattern:
            in_pattern, out_pattern = pattern.split("->")
            in_dims = translate_pattern(in_pattern)
        else:
            out_pattern = pattern
            in_dims = None
        out_dims = translate_pattern(out_pattern)
        return _rearrange(da, out_dims=out_dims, in_dims=in_dims, **kwargs)
    return _rearrange(da, out_dims=pattern, in_dims=pattern_in, **kwargs)


def _reduce(da, reduction, out_dims, in_dims=None, **kwargs):
    """Wrap `einops.reduce <https://einops.rocks/api/reduce/>`_.

    This is the function that actually interfaces with ``einops``.
    :func:`xarray_einstats.einops.rearrange` is the user facing
    version as it exposes two possible APIs, one of them significantly
    less verbose and more friendly (but much less flexible).

    Parameters
    ----------
    da : xarray.DataArray
        Input DataArray to be reduced
    reduction : string or callable
        One of available reductions ('min', 'max', 'sum', 'mean', 'prod') by ``einops.reduce``,
        case-sensitive. Alternatively, a callable ``f(tensor, reduced_axes) -> tensor``
        can be provided. ``reduced_axes`` are passed as a list of int.
    out_dims : list of str, list or dict
        The output pattern for the dimensions.
        The dimensions present in
    in_dims : list of str or dict, optional
        The input pattern for the dimensions.
        This is only necessary if you want to split some dimensions.
    kwargs : dict, optional
        kwargs with key equal to dimension names in ``out_dims``
        (that is, strings or dict keys) are passed to einops.rearrange
        the rest of keys are passed to :func:`xarray.apply_ufunc`
    """
    da_dims = da.dims

    handler = DimHandler()
    if in_dims is None:
        in_dims = []
        in_names = []
        in_pattern = ""
    else:
        in_dims, in_names, in_pattern = process_pattern_list(
            in_dims, handler=handler, allow_list=False
        )
    # note, not using sets for da_dims to avoid transpositions on missing variables,
    # if they wanted to transpose those they would not be missing variables
    out_dims, out_names, out_pattern = process_pattern_list(out_dims, handler=handler)
    missing_in_dims = [dim for dim in da_dims if dim not in in_names]
    pattern = f"{handler.get_names(missing_in_dims)} {in_pattern} -> {out_pattern}"

    all_dims = set(out_dims + out_names + in_names + in_dims)
    axes_lengths = {handler.rename_kwarg(k): v for k, v in kwargs.items() if k in all_dims}
    kwargs = {k: v for k, v in kwargs.items() if k not in all_dims}
    return xr.apply_ufunc(
        einops.reduce,
        da,
        pattern,
        reduction,
        input_core_dims=[missing_in_dims + in_names, [], []],
        output_core_dims=[out_names],
        kwargs=axes_lengths,
        **kwargs,
    )


def reduce(da, pattern, reduction, pattern_in=None, **kwargs):
    """Expose `einops.reduce <https://einops.rocks/api/reduce/>`_ with an xarray-like API.

    It has two possible syntaxes which are independent and somewhat complementary.

    Parameters
    ----------
    da : xarray.DataArray
        Input array
    pattern : str or list of [str, list or dict]
        If `pattern` is a string, it uses the same syntax as einops
        with two caveats:

        * Unless splitting or stacking, you must use the actual dimension names.
        * When splitting or stacking you can use ``(dim1 dim2)=dim``. This is *necessary*
          for the left hand side as it identifies the dimension to split, and
          optional on the right hand side, if omitted the stacked dimension will be given
          a default name.

        If `pattern` is not a string, then it must be a list where each of its elements
        is one of: ``str``, ``list`` (to stack those dimensions and give them an
        arbitrary name) or ``dict of {str: list}`` (to stack the dimensions indicated
        as values of the dictionary and name the resulting dimensions with the key).

        `pattern` is then interpreted as the output side of the einops pattern. See
        TODO for more details.
    reduction : string or callable
        One of available reductions ('min', 'max', 'sum', 'mean', 'prod') by ``einops.reduce``,
        case-sensitive. Alternatively, a callable ``f(tensor, reduced_axes) -> tensor``
        can be provided. ``reduced_axes`` are passed as a list of int.
    pattern_in : list of [str or dict], optional
        The input pattern for the dimensions. It can only be provided if `pattern`
        is a ``list``. Also, note this is only necessary if you want to split some dimensions.

        The syntax and interpretation is the same as the case when `pattern` is a list,
        with the only difference that ``list`` elements are not allowed, the same way
        that ``(dim1 dim2)=dim`` is required on the left hand side when using string
    kwargs : dict, optional
        Passed to :func:`xarray_einstats.einops.reduce`

    Returns
    -------
    xarray.DataArray

    Notes
    -----
    Unlike for general xarray objects, where dimension
    names can be :term:`hashable <xarray:name>` here
    dimension names are not recommended but required to be
    strings for both cases. Future releases however might
    support this when using lists as `pattern`, comment
    on :issue:`50` if you are interested in the feature
    or could help implement it.

    See Also
    --------
    xarray_einstats.einops.rearrange
    """
    if isinstance(pattern, str):
        if "->" in pattern:
            in_pattern, out_pattern = pattern.split("->")
            in_dims = translate_pattern(in_pattern)
        else:
            out_pattern = pattern
            in_dims = None
        out_dims = translate_pattern(out_pattern)
        return _reduce(da, reduction, out_dims=out_dims, in_dims=in_dims, **kwargs)
    return _reduce(da, reduction, out_dims=pattern, in_dims=pattern_in, **kwargs)


def raw_reduce(*args, **kwargs):
    """Wrap einops.reduce.

    DEPRECATED
    """
    warnings.warn(
        "raw_reduce has been deprecated. Its functionality has been merged into reduce",
        DeprecationWarning,
    )
    return reduce(*args, **kwargs)


def raw_rearrange(*args, **kwargs):
    """Wrap einops.rearrange.

    DEPRECATED
    """
    warnings.warn(
        "raw_rearrange has been deprecated. Its functionality has been merged into rearrange",
        DeprecationWarning,
    )
    return rearrange(*args, **kwargs)


class DaskBackend(einops._backends.AbstractBackend):  # pylint: disable=protected-access
    """Dask backend class for einops.

    It should be imported before using functions of :mod:`xarray_einstats.einops`
    on Dask backed DataArrays.
    It doesn't need to be initialized or used explicitly

    Notes
    -----
    Class created from the advise on
    `issue einops#120 <https://github.com/arogozhnikov/einops/issues/120>`_ about Dask support.
    And from reading
    `einops/_backends <https://github.com/arogozhnikov/einops/blob/master/einops/_backends.py>`_,
    the source of the AbstractBackend class of which DaskBackend is a subclass.
    """

    # pylint: disable=no-self-use
    framework_name = "dask"

    def __init__(self):
        """Initialize DaskBackend.

        Contains the imports to avoid errors when dask is not installed
        """
        import dask.array as dsar

        self.dsar = dsar

    def is_appropriate_type(self, tensor):
        """Recognizes tensors it can handle."""
        return isinstance(tensor, self.dsar.core.Array)

    def from_numpy(self, x):  # noqa: D102
        return self.dsar.array(x)

    def to_numpy(self, x):  # noqa: D102
        return x.compute()

    def arange(self, start, stop):  # noqa: D102
        # supplementary method used only in testing, so should implement CPU version
        return self.dsar.arange(start, stop)

    def stack_on_zeroth_dimension(self, tensors: list):  # noqa: D102
        return self.dsar.stack(tensors)

    def tile(self, x, repeats):  # noqa: D102
        return self.dsar.tile(x, repeats)

    def is_float_type(self, x):  # noqa: D102
        return x.dtype in ("float16", "float32", "float64", "float128")

    def add_axis(self, x, new_position):  # noqa: D102
        return self.dsar.expand_dims(x, new_position)
